// S88 DI board
// Copyright(C)2016- Desktop Station
// reference : RB_S88, Fujigaya's S88 Control board
// 
//160512 Yaasan    Improve coding rule and add S88 DI features

#include <EEPROM.h>

#define		PORT_DATAIN		5
#define		PORT_DATAOUT	6
#define		PORT_BTNA		A0
#define		PORT_BTNB		A1
#define		PORT_LED1		8
#define		PORT_LED2		9
#define		PORT_LED3		10
#define		PORT_LED4		11
#define		PORT_LED5		A2
#define		PORT_LED6		A3
#define		PORT_LED7		A4
#define		PORT_LED8		A5

#define		BTNA_SW1		0b00000001
#define		BTNA_SW2		0b00000010
#define		BTNA_SW3		0b00000100
#define		BTNA_SW4		0b00001000
#define		BTNA_SW5		0b00010000
#define		BTNA_SW6		0b00000001
#define		BTNA_SW7		0b00000010
#define		BTNA_SW8		0b00000100
#define		BTNA_SW9		0b00001000
#define		BTNA_SW10		0b00010000

#define		STATE_RUN		1
#define		STATE_CFG		10


//Sensor control
#define		 INTERVAL_L5	100 //L5 interval[ms]
#define		 MAX_S88PORTNUM	8 //port_num to see from DCC Controller


/* RFID 64bit data  
   example 1500 6850 3B16
*/

byte gTempRcvIdx = 0;
byte gTempRcvData[14];
byte gRFIDtable[8][12];

//Configuration mode button
byte gConfigSlotNo = 255;

//State Diagram
byte gState = STATE_RUN;

byte gDispCnt = 0;


byte gClockCounter = 0;
unsigned long gPreviousL5 = 0;
unsigned long gPreviousL6 = 0;

//Logic IC emulation buffer
unsigned long gTransferData = 0; // same as shift register value
word gSensorData = 0;   // same as flip flop buffer
byte gStatusData[8] = {0,0,0,0,0,0,0,0};


//Button buffer
byte gBtnDataA = 0;
byte gBtnDataB = 0;


/* Functions */
void ControlLEDs();
void ProcessCheck(byte inBufA, byte inBufB, byte inTempA, byte inTempB);
byte GetButtonsFromAD(unsigned short inADData);
void MonitorButtons();

/***********************************************************************
 * 
 *  Functions
 * 
***********************************************************************/

void setup() 
{
	
	Serial.begin(9600);
	while (!Serial);

	Serial.println("--------------------------------------");
	Serial.println("Desktop Station RFID S88 board V0.1     ");
	Serial.println("--------------------------------------");
	Serial.println("100 Ready");	
	
	//S88 Control
	attachInterrupt(0, irqCLK, RISING); //pin 2 = clock interrupt
	attachInterrupt(1, irqPS, RISING);  //pin 3 = PS interrupt
	pinMode(2, INPUT_PULLUP);
	pinMode(3, INPUT_PULLUP);
	pinMode(PORT_DATAIN, INPUT_PULLUP); // data in from S88 in chain
	pinMode(PORT_DATAOUT, OUTPUT); // data out to commnad station or to previous S88 chain
	digitalWrite(PORT_DATAOUT, LOW);   

	//Buttons
	pinMode(PORT_BTNA, INPUT);
	pinMode(PORT_BTNB, INPUT);

	//LEDs
	pinMode(PORT_LED1, OUTPUT);
	pinMode(PORT_LED2, OUTPUT);
	pinMode(PORT_LED3, OUTPUT);
	pinMode(PORT_LED4, OUTPUT);
	pinMode(PORT_LED5, OUTPUT);
	pinMode(PORT_LED6, OUTPUT);
	pinMode(PORT_LED7, OUTPUT);
	pinMode(PORT_LED8, OUTPUT);
	
	//EEPROMǂݏo
	LoadEEPROM();
	
	//Task base
	gPreviousL5 = millis();
	gPreviousL6 = millis();

	//init sensor_value,transfer_value
	gTransferData = 0;
	gSensorData = 0;

}

/* Main loop */
void loop()
{
	word aSensorData = 0;
	
	//Serial check
	byte aRFIDData[12];
	
	if( ReadRDMFrame(aRFIDData) > 0)
	{
		
		switch(gState)
		{
			
		case STATE_RUN:
			//Check Table
			
			for( int i = 0; i < 8; i++)
			{
				
				if( CheckRFIDData(aRFIDData, gRFIDtable[i]) > 0)
				{
					gStatusData[i] = 20;//2sec
				}
				else
				{
					
				}
				
			}
			
			if( aSensorData > 0)
			{
				Serial.print("Detected >>");
				Serial.print(aSensorData, BIN);
				Serial.println("");
			}
			
			
			break;
			
		case STATE_CFG:
			gRFIDtable[gConfigSlotNo][0] = aRFIDData[0];
			gRFIDtable[gConfigSlotNo][1] = aRFIDData[1];
			gRFIDtable[gConfigSlotNo][2] = aRFIDData[2];
			gRFIDtable[gConfigSlotNo][3] = aRFIDData[3];
			gRFIDtable[gConfigSlotNo][4] = aRFIDData[4];
			gRFIDtable[gConfigSlotNo][5] = aRFIDData[5];
			gRFIDtable[gConfigSlotNo][6] = aRFIDData[6];
			gRFIDtable[gConfigSlotNo][7] = aRFIDData[7];
			gRFIDtable[gConfigSlotNo][8] = aRFIDData[8];
			gRFIDtable[gConfigSlotNo][9] = aRFIDData[9];
			gRFIDtable[gConfigSlotNo][10] = aRFIDData[10];
			gRFIDtable[gConfigSlotNo][11] = aRFIDData[11];

			gState = STATE_RUN;
			
			//EEPROMۑ
			SaveEEPROM(gConfigSlotNo);
			
			Serial.print("Registered No.");
			Serial.print(gConfigSlotNo);
			Serial.println("");
			break;
		
		}
	}
	else
	{
	}
	
	//Interval 100ms
	if(millis() - gPreviousL5 > INTERVAL_L5)
	{
		
		//Refresh Sensor Data
		switch(gState)
		{
		case STATE_RUN:
			
			for( int i = 0; i < 8; i++)
			{
				if( gStatusData[i] > 0)
				{
					gStatusData[i]--;
				}
				else
				{
					gStatusData[i] = 0;
				}
				
				aSensorData = aSensorData | (((gStatusData[i] > 0) ? 1 : 0) << i);
			}
			
			//S88f[^XV
			gSensorData = aSensorData;
			
			break;
			
		}		
		
		
		MonitorButtons();
		
		ControlLEDs();
		
		gPreviousL5 = millis();
		

	}
	
	if(millis() - gPreviousL6 > 333)
	{	
		gPreviousL6 = millis();
		
		//ݒԕ\؂ւpJE^
		gDispCnt++;
		
		if( gDispCnt > 1)
		{
			gDispCnt = 0;
		}		
	}
	
}

void irqPS()
{
	//Set transfer value
	gClockCounter = 0;
	
	//Set Shift data
	gTransferData = gSensorData;
	
}

void irqCLK()
{
	//transfer anf emulate shift register.
	
	//Write transfer_value
	digitalWrite(PORT_DATAOUT, bitRead(gTransferData, gClockCounter));
	
	//Delay (e.g. 20ms)
	delayMicroseconds(16); 
	
	//Store PORT_DATAIN value
	bitWrite(gTransferData, gClockCounter, digitalRead(PORT_DATAIN));
	
	//Increment clock
	gClockCounter++;
	
	if( gClockCounter >= MAX_S88PORTNUM)
	{
		gClockCounter = 0;
	}
}


void ControlLEDs()
{
	byte aLEDPower = 0;
	
	for( int i = 0; i < 8; i++)
	{
		
		aLEDPower = gStatusData[i];
		
		if( (i == gConfigSlotNo) && (gState == STATE_CFG))
		{
			if( gDispCnt > 0)
			{
				aLEDPower = 1;
				
			}
			else
			{
				aLEDPower = 0;
				
			}
		}
		
		switch(i)
		{
		case 0:
			digitalWrite(PORT_LED1, aLEDPower);
			break;
		case 1:
			digitalWrite(PORT_LED2, aLEDPower);
			break;
		case 2:
			digitalWrite(PORT_LED3, aLEDPower);
			break;
		case 3:
			digitalWrite(PORT_LED4, aLEDPower);
			break;
		case 4:
			digitalWrite(PORT_LED5, aLEDPower);
			break;
		case 5:
			digitalWrite(PORT_LED6, aLEDPower);
			break;
		case 6:
			digitalWrite(PORT_LED7, aLEDPower);
			break;
		case 7:
			digitalWrite(PORT_LED8, aLEDPower);
			break;
		}
	}
}

void MonitorButtons()
{
	byte aTempBtnA = 0;
	byte aTempBtnB = 0;
	
	unsigned short aButtonA = analogRead(PORT_BTNA);
	unsigned short aButtonB = analogRead(PORT_BTNB);
	
	//Get button status
	aTempBtnA = GetButtonsFromAD(aButtonA);
	aTempBtnB = GetButtonsFromAD(aButtonB);
	
	//Check and process the buttons
	ProcessCheck(gBtnDataA, gBtnDataB, aTempBtnA, aTempBtnB);
	
	//Store to buffer
	gBtnDataA = aTempBtnA;
	gBtnDataB = aTempBtnB;
	
}

void ProcessCheck(byte inBufA, byte inBufB, byte inTempA, byte inTempB)
{
	byte aChanged = 0;
	
	if( (inBufA == 0) && (inTempA > 0))
	{
		
		switch(inTempA)
		{
		case 1:
			//SW1
			gConfigSlotNo = 0;
			break;
		case 2:
			//SW2
			gConfigSlotNo = 1;
			break;
		case 4:
			//SW3
			gConfigSlotNo = 2;
			break;
		case 8:
			//SW4
			gConfigSlotNo = 3;
			break;
		case 16:
			//SW5
			gConfigSlotNo = 4;
			break;
		}
		
		aChanged++;
		
	}
	
	if( (inBufB == 0) && (inTempB > 0))
	{
		
		switch(inTempB)
		{
		case 1:
			//SW6
			gConfigSlotNo = 5;
			break;
		case 2:
			//SW7
			gConfigSlotNo = 6;
			break;
		case 4:
			//SW8
			gConfigSlotNo = 7;
			break;
		case 8:
			//SW9
			//gSensorData = 0;
			break;
		case 16:
			//SW10
			//gSensorData = 0b11111111;
			break;
		}
		
		aChanged++;
	}
	
	if( aChanged > 0)
	{
		//܂ꂽ
		if( gState == STATE_CFG)
		{
			//ʏ탂[h
			gState = STATE_RUN;
		}
		else
		{
			//ݒ胂[h
			gState = STATE_CFG;
		}
		
		Serial.print("BUTTON: ");
		Serial.println(gSensorData);
	}
	
}

byte GetButtonsFromAD(unsigned short inADData)
{
	/*
	R1	2.2k
	R2	470
	R3	1k
	R4	2.2k
	R5	6.8k
	
	Threshold value:
	SW1: 82
	SW2: 290
	SW3: 525
	SW4: 747
	SW5: 942
	
	
	*/
	
	byte aRet = 0;
	
	if( inADData <= 82)
	{
		aRet = 1 << 0;
	}
	else if( inADData <= 290)
	{
		aRet = 1 << 1;
	}
	else if( inADData <= 525)
	{
		aRet = 1 << 2;
	}
	else if( inADData <= 747)
	{
		aRet = 1 << 3;
	}
	else if( inADData <= 942)
	{
		aRet = 1 << 4;
	}
	else
	{
		//Not pressed
	}
	
	
	return aRet;
}

byte ReadRDMFrame(byte outpReadData[])
{
	byte aByte;
	byte aRes = 0;
	
	//Serial check
	if( Serial.available() > 0)
	{	
		aByte = Serial.read();
		
		//Header check
		if( aByte == 2)
		{
			gTempRcvData[0] = 0x02;
			gTempRcvIdx = 1;
		}
		else
		{
			gTempRcvData[gTempRcvIdx] = aByte;
			gTempRcvIdx++;
			
			if( gTempRcvIdx >= 14)
			{
				
				if( gTempRcvData[13] == 3)
				{
					
					Serial.print("RFID>>");
					Serial.write(gTempRcvData[1]);
					Serial.write(gTempRcvData[2]);
					Serial.write(gTempRcvData[3]);
					Serial.write(gTempRcvData[4]);
					Serial.write(gTempRcvData[5]);
					Serial.write(gTempRcvData[6]);
					Serial.write(gTempRcvData[7]);
					Serial.write(gTempRcvData[8]);
					Serial.write(gTempRcvData[9]);
					Serial.write(gTempRcvData[10]);
					Serial.write(gTempRcvData[11]);
					Serial.write(gTempRcvData[12]);
					Serial.println("");
					
					//Valid RFID frame
					outpReadData[0] = gTempRcvData[1];
					outpReadData[1] = gTempRcvData[2];
					outpReadData[2] = gTempRcvData[3];
					outpReadData[3] = gTempRcvData[4];
					outpReadData[4] = gTempRcvData[5];
					outpReadData[5] = gTempRcvData[6];
					outpReadData[6] = gTempRcvData[7];
					outpReadData[7] = gTempRcvData[8];
					outpReadData[8] = gTempRcvData[9];
					outpReadData[9] = gTempRcvData[10];
					outpReadData[10] = gTempRcvData[11];
					outpReadData[11] = gTempRcvData[12];
					
					//Execute RFID Compare
					aRes = 1;
				}
				
				//Clear serial buffer
				Serial.flush();
				
				gTempRcvIdx = 0;
				
			}
		}
	}
	
	return aRes;
}

byte CheckRFIDData(byte inpReadData[], byte inpTableData[])
{
	byte aRes = 0;
	byte aCorrectCnt = 0;
	
	for( int i = 0; i < 12; i++)
	{
		if( inpReadData[i] == inpTableData[i])
		{
			aCorrectCnt++;
		}
	}
	
	if( aCorrectCnt == 12)
	{
		aRes = 1;
		
	}
	
	return aRes;
}


void LoadEEPROM()
{
	
	if( EEPROM.read(0) == 0xBB)
	{
		
		for( int i = 0; i < 8; i++)
		{
			for( int j = 0; j < 12; j++)
			{
				gRFIDtable[i][j] = EEPROM.read( 12 + (i * 12) + j);
				
			}
		}
	}	
	
}

void SaveEEPROM(byte inMode)
{
	
	if( EEPROM.read(0) == 0xBB)
	{
		
		for( int j = 0; j < 12; j++)
		{
			EEPROM.write( 12 + (inMode * 12) + j, gRFIDtable[inMode][j]);
		}
		
	}
	else
	{
		//
		EEPROM.write(0, 0xBB);
		
		//
		for( int i = 0; i < 8; i++)
		{
			for( int j = 0; j < 12; j++)
			{
				EEPROM.write( 12 + (i * 12) + j, gRFIDtable[i][j]);
				
			}
		}
	}
	
}
